// AutoCapacitanceMeter program version 2a. For Arduino 0018 or newer.
// (c)20110312 by MODAT7 under GNU GPLv1. Original single code by Paul Badger.
// This was written for an ATmega328 but may work with an ATmega168 (code is about 11k compiled).

/* Auto-Ranging Notes.

When wiring up the hardware, the resistors should go from smallest value to largest value.

Even though some delays are programmed in, the first resistor doesn't always start the reading.
If the reading time takes too long, just unplug the capacitor and plug it back in again.
Even with the checks and delays, often times the first reading is significantly off. Wait
for the second reading to come around.

Averaging can be implemented by making multiple Get() calls and keeping track of the milliFarad
variable. Note that this will greatly slow down readings on larger capacitors.

For decent auto-ranging features, it is recommended to have 8 resistors starting at 560 ohms (or
preferrably 1k) and increasing the size by doubling (or more) to about 3meg-ohms.
Recommended values to pick from. Higher values will meter smaller capacitor sizes.
560, 1k, 2.2k, 10k, 22k, 56k, 100k, 560k, 1m, 3.3m
(560k and 1m were swapped for a single 750k in my prototype for 9 resistors total.)
Available pins: 11 digital pins for resistors (including drain), 5 ADC pins in digital mode.
Tweaking the resistor value up in the code will lower the calculated capacitance.
A 10k resistor will take about 10 seconds to read 1 milliFarad.

Observations. Lower milliamp charging seems to ignore ESR issues more. I paralleled 5x 220uF's
and got 17uF higher on the reading. The 560 charge resistor seem to always give a slightly higher
reading. 5x 220uF's in series gives a close to expected number for my 19k76 resistor. For my
1083R, the value is lower (ESR issue?).

For version 2a, I added a standard 16x2 parallel interfaced LCD in 4 bit mode for standalone
operation. This will need an additional 6 pins from the AVR. I put the 5 extra ADC's in digital
mode so I could have a total of 9 charge resistors for auto-ranging.
*/

/* Paul Badger's original notes:
 *
 *  RCTiming_capacitance_meter
 *   Paul Badger 2008
 *  Demonstrates use of RC time constants to measure the value of a capacitor 
 *
 * Theory   A capcitor will charge, through a resistor, in one time constant, defined as T seconds where
 *    TC = R * C
 * 
 *    TC = time constant period in seconds
 *    R = resistance in ohms
 *    C = capacitance in farads (1 microfarad (ufd) = .0000001 farad = 10^-6 farads ) 
 *
 *    The capacitor's voltage at one time constant is defined as 63.2% of the charging voltage.
 *
 *  Hardware setup:
 *  Test Capacitor between common point and ground (positive side of an electrolytic capacitor  to common)
 *  Test Resistor between chargePin and common point
 *  220 ohm resistor between dischargePin and common point
 *  Wire between common point and analogPin (A/D input) 
 */


#include "CapacitanceMeter.h"
#include <LiquidCrystal.h>
#include <stdio.h> //sprintf

#define SERIALBAUD 115200 //Faster serial means shorter delays in the program.

#define dPinADCCapacitor 0 //ADC pin for measuring capacitor voltage.
#define dCapacitorADCValue 647 // 647 is 63.2% of 1023 (at 10bits), which corresponds to full-scale voltage.
#define dPinLED 13 //shows charging status/activity.
#define dPinDischargeResistor 11 //used to drain the capacitor after charging.
#define dCapacitorMinTime 200 //milliseconds, below this, try the next resistor size up for accuracy.
#define dCapacitorAbsMinTime 20 //milliseconds, times below this are very inaccurate and meaningless.

//Define each object per charge resistor.
//I use DP pins 2-10 below.
CapacitanceMeter acm1, acm2, acm3, acm4, acm5, acm6, acm7, acm8, acm9;

//Set up the LCD.
//The #define's make it clearer to change the pins/settings around if needed.
#define dLCDPinRS 12
#define dLCDPinEN 15
#define dLCDPinD4 16
#define dLCDPinD5 17
#define dLCDPinD6 18
#define dLCDPinD7 19
#define dLCDColumns 16 //for a 16x2 LCD.
#define dLCDRows 2 //must be 2 rows for now.

LiquidCrystal lcd(dLCDPinRS, dLCDPinEN, dLCDPinD4, dLCDPinD5, dLCDPinD6, dLCDPinD7);

//used for sprintf
char LCDLineBuffer[(dLCDColumns*2)+1]; //give it some space.
unsigned char LCDRowPosition;



//LCD clear line function.
void LCDClearLine(unsigned char LineNumber) {
if (LineNumber < 0 || LineNumber >= dLCDRows) {return;} //don't write outside the LCD.
lcd.setCursor((dLCDColumns-1), LineNumber);
for(int i=(dLCDColumns-1); i>=0; i--) {
  lcd.setCursor(i, LineNumber);
  lcd.write(' ');
  }
lcd.setCursor(0, LineNumber);
//This should leave the cursor at column 0 ready to print on the line.
} //end LCDClearLine()



//LCD print function.
//If the LCD is using ADC pins as digital pins, call this early so ADC has time to resettle.
void LCDPrint(CapacitanceMeter acm, unsigned char LineNumber) {
if(acm.milliFarads==0.0) {return;} //don't print out empty values.
LCDRowPosition = sprintf(LCDLineBuffer, "%d ", (LineNumber+1) );
//Float linking is missing, so the uglier modulus math is used instead.
//Modulus math will be a little different than float and will have slightly different results than printed serial.
if(acm.milliFarads < 1.0e-3) { //print the value out as nanoFarads.
  LCDRowPosition += sprintf(LCDLineBuffer+LCDRowPosition, "%d.%02dnF",
    ((int)(acm.milliFarads*1.0e6)),
    ( ((int)(acm.milliFarads*1.0e6*100))%100 ) );
  }
else if(acm.milliFarads < 1.0) { //print the value out as microFarads.
  LCDRowPosition += sprintf(LCDLineBuffer+LCDRowPosition, "%d.%02duF",
    ((int)(acm.milliFarads*1.0e3)),
    ( ((int)(acm.milliFarads*1.0e3*100))%100 ) );
  }
else { //print the value out as milliFarads.
  LCDRowPosition += sprintf(LCDLineBuffer+LCDRowPosition, "%d.%03dmF",
    ((int)(acm.milliFarads)),
    ( ((int)(acm.milliFarads*1000))%1000 ) );
  }

LCDClearLine(LineNumber);
if(LCDRowPosition > dLCDColumns) {
  //add a little safety.
  lcd.print("PRT OF"); // print overflow.
  return;
  }
lcd.print(LCDLineBuffer);
//Serial.println(LCDLineBuffer); //debug
} //end LCDPrint()



//Do the math to show percent difference between resistor levels.
//This will hint at ESR. Higher percentages are worse. Below 1% seems to be good.
//If the resistors aren't precisely metered, these values will be skewed.
//Comparing the two values, the resistors should be at least double/half of each other.
//This math needs to be normalized and expanded a little more.
//have: ChargeResistorValue, TimeElapsed, milliFarads
void CapacitorESRPercent(double mf1, double mf2) {
static double mftemp;
//Keep the percentage greater than 1.
mftemp = mf1 / mf2;
if(mftemp<1.0) {mftemp = mf2 / mf1;}
//convert to whole number percent
mftemp = (mftemp-1) * 100;
//show the results.
//Since the float library doesn't get linked, modulus math will be slightly different than serial.
sprintf(LCDLineBuffer, " %d.%02d%%", ((int)mftemp), ( ((int)(mftemp*100))%100 ) );
lcd.print(LCDLineBuffer);
Serial.print("ESR diff %: ");
Serial.println(mftemp, 2 );
} //end CapacitorESRPercent()



//Check to see if a capacitor is connected that is in range.
//Do one object for each charge resistor.
//Resistor sizes should be from smallest to largest.
boolean CapacitorCheck() {
acm1.Check(); if(acm1.TimeStart>0) {return true;}
acm2.Check(); if(acm2.TimeStart>0) {return true;}
acm3.Check(); if(acm3.TimeStart>0) {return true;}
acm4.Check(); if(acm4.TimeStart>0) {return true;}
acm5.Check(); if(acm5.TimeStart>0) {return true;}
acm6.Check(); if(acm6.TimeStart>0) {return true;}
acm7.Check(); if(acm7.TimeStart>0) {return true;}
acm8.Check(); if(acm8.TimeStart>0) {return true;}
acm9.Check(); if(acm9.TimeStart>0) {return true;}
return false; //nothing was found.
} //end CapacitorCheck()



//Do the capacitor measuring and report.
//These stacked time tests implement the auto-ranging feature.
//Do one object block for each acm/charge resistor. Be careful of the numbering.
//Resistor sizes should be from smallest to largest.
void CapacitorGet() {
static unsigned char GetCount;

GetCount=0; //used for the ESR difference percentage.

//First acm Get().
//Rename the object if needed, but don't mess with this block (being the first).
acm1.Get();
if(acm1.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm1, GetCount);
  acm1.SerialPrint();
  GetCount = 1;
  }

//Middle acm object. Use this as a middle template for more objects.
//Rename if objects if needed.
acm2.Get();
if(acm2.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm2, GetCount);
  acm2.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm1.milliFarads, acm2.milliFarads);
    return;
    }
  }

acm3.Get();
if(acm3.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm3, GetCount);
  acm3.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm2.milliFarads, acm3.milliFarads);
    return;
    }
  }

acm4.Get();
if(acm4.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm4, GetCount);
  acm4.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm3.milliFarads, acm4.milliFarads);
    return;
    }
  }

acm5.Get();
if(acm5.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm5, GetCount);
  acm5.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm4.milliFarads, acm5.milliFarads);
    return;
    }
  }

acm6.Get();
if(acm6.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm6, GetCount);
  acm6.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm5.milliFarads, acm6.milliFarads);
    return;
    }
  }

acm7.Get();
if(acm7.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm7, GetCount);
  acm7.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm6.milliFarads, acm7.milliFarads);
    return;
    }
  }

acm8.Get();
if(acm8.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm8, GetCount);
  acm8.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm7.milliFarads, acm8.milliFarads);
    return;
    }
  }

//Last acm Get().
//Rename objects if needed, but don't mess with this block (being the last).
acm9.Get();
if(acm9.TimeElapsed >= dCapacitorMinTime) {
  LCDPrint(acm9, GetCount);
  if(GetCount==0) {  //clear any leftovers and print warning.
    LCDClearLine(1); lcd.print("2 End Of Range");
    }
  acm9.SerialPrint();
  GetCount += 1;
  if(GetCount >= 2) {
    CapacitorESRPercent(acm8.milliFarads, acm9.milliFarads);
    }
  return;
  }

//Make this the last "acm" object.
//Any times left will be below dCapacitorMinTime.
if(acm9.TimeElapsed < dCapacitorAbsMinTime) {
  //If time is too short, print out below range error.
  //Times below this have a very poor accuracy.
  lcd.clear();
  lcd.print("Below Range");
  Serial.println("Below Range");
  }
else {
  //...else print out something but with a low accuracy warning.
  LCDPrint(acm9, GetCount);
  lcd.print(" *LOW");
  LCDClearLine(1); lcd.print("2 End Of Range");
  Serial.print("*LOW ");
  acm9.SerialPrint();
  }

} //end CapacitorGet()



void setup() {

//pin floating paranoia, set pullup's (don't mess with pins 0 and 1 serial).
for(int i=2; i<20; i++) {
  pinMode(i, INPUT);
  digitalWrite(i, HIGH);
  }

//initialize serial transmission for output.
Serial.begin(SERIALBAUD);

//set up the LCD.
pinMode(dLCDPinRS, OUTPUT); digitalWrite(dLCDPinRS, LOW);
pinMode(dLCDPinEN, OUTPUT); digitalWrite(dLCDPinEN, LOW);
pinMode(dLCDPinD4, OUTPUT); digitalWrite(dLCDPinD4, LOW);
pinMode(dLCDPinD5, OUTPUT); digitalWrite(dLCDPinD5, LOW);
pinMode(dLCDPinD6, OUTPUT); digitalWrite(dLCDPinD6, LOW);
pinMode(dLCDPinD7, OUTPUT); digitalWrite(dLCDPinD7, LOW);
lcd.begin(dLCDColumns, dLCDRows);
lcd.clear(); //my LCD displays trash at turn on.
lcd.blink(); //turn on cursor blink.

//clear LCDLineBuffer:
LCDLineBuffer[0]='\0';
LCDRowPosition=0;

//Set up the pins and values.
//Note that the number of charge resistors (acm objects) used will have to be accounted for here
//and in the functions above.
//template:
/* CapacitanceMeter::SetUp(unsigned char sPinChargeResistor, unsigned char sPinADCCapacitor,
   unsigned char sPinDischargeResistor, unsigned char sPinLED, unsigned int sCapacitorADCValue,
   double sChargeResistorValue) */
acm1.SetUp(2, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 501.0);
acm2.SetUp(3, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 1083.0);
acm3.SetUp(4, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 2170.0);
acm4.SetUp(5, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 9710.0);
acm5.SetUp(6, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 19760.0);
acm6.SetUp(7, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 54700.0);
acm7.SetUp(8, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 99800.0);
acm8.SetUp(9, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 776000.0);
acm9.SetUp(10, dPinADCCapacitor, dPinDischargeResistor, dPinLED, dCapacitorADCValue, 3060000.0);

} //end setup()



void loop() {

if(CapacitorCheck()) {
  delay(200); //give the connectors a little time to settle to avoid false readings.
  CapacitorGet(); //auto-ranging function.
  Serial.println(""); //space it out for easier reading.
  }

} //end loop()

