// CapacitanceMeter.cpp class code. For Arduino 0018 or newer.
// CapacitanceMeter class to consolidate various variables and functions.
// (c)20110312 by MODAT7 under GNU GPLv1. Original single code by Paul Badger.

#include "CapacitanceMeter.h"

//The classes would normally use read() and set() commands and "private" the data,
//but space is at a premium here in an MCU. This has been written for flexibility.
//"this->" allows for better clarity and less confusion if global variables have
//the same name as class variables.



//constructor. This could be more but I want to code for size small.
/* Space is at a premium and this isn't necessary since it is taken care of in SetUp().
CapacitanceMeter::CapacitanceMeter() {
this->PinChargeResistor=9;
this->PinADCCapacitor=0;
this->PinDischargeResistor=8;
this->PinLED=13;
this->CapacitorADCValue=647;
this->ChargeResistorValue=10000.0;
}
*/



//Set up the ADC and IO pins.
void CapacitanceMeter::SetUp(unsigned char sPinChargeResistor, unsigned char sPinADCCapacitor,
   unsigned char sPinDischargeResistor, unsigned char sPinLED, unsigned int sCapacitorADCValue,
   double sChargeResistorValue) {
//Set up charge resistor pin.
this->PinChargeResistor=sPinChargeResistor;
pinMode(this->PinChargeResistor, INPUT); //set to input (high impedance)
digitalWrite(this->PinChargeResistor, LOW); //turn off pullup.

//set up ADC pin. Take the pin out of pull-up mode.
//Arduino analog pins are 14-19 when accessed in digital mode.
this->PinADCCapacitor=sPinADCCapacitor;
pinMode((14 + this->PinADCCapacitor), INPUT);
digitalWrite((14 + this->PinADCCapacitor), LOW);
//The first analog read of a pin may be bad, so read and discard it.
analogRead(this->PinADCCapacitor);

//set up the discharge pin.
this->PinDischargeResistor=sPinDischargeResistor;
pinMode(this->PinDischargeResistor, INPUT); //set to input (high impedance)
digitalWrite(this->PinDischargeResistor, LOW); //turn off pullup.

//set up the LED pin, default off.
this->PinLED=sPinLED;
pinMode(this->PinLED, OUTPUT);
digitalWrite(this->PinLED, LOW);

//copy over final vars.
this->CapacitorADCValue=sCapacitorADCValue;
this->ChargeResistorValue=sChargeResistorValue;
}



//Check to see if the capacitor is present. This is intended to be relatively quick.
void CapacitanceMeter::Check() {
  pinMode(this->PinChargeResistor, OUTPUT); //set to output (low impedance).
  //for fine tuning, wait until a millisecond tick.
  digitalWrite(this->PinLED, HIGH); //set status LED to on.
  this->TimeStart=millis()+2;
  while(millis() < this->TimeStart) {} //wait for the tick.
  digitalWrite(this->PinChargeResistor, HIGH); //set pin HIGH for capacitor charging.
  this->TimeStart = millis(); //start the clock.
  while(analogRead(this->PinADCCapacitor) < this->CapacitorADCValue) {
    this->TimeElapsed = millis() - this->TimeStart;
    if(this->TimeElapsed>0) {break;} //a capacitor was found, no need to continue.
    }
  digitalWrite(this->PinLED, LOW); //set status LED to off.
  this->Discharge(7); //drain and set up for next run. Small number means quick drain mode.
  this->TimeStart = this->TimeElapsed; //TimeElapsed is used for other things.
  //use "TimeStart>0" to see if a capacitor is attached.
} //end Check();



//Get the full time value on the resistor pin.
//This is timing critical so code cannot be shared/overloaded.
void CapacitanceMeter::Time() {
  pinMode(this->PinChargeResistor, OUTPUT); //set to output (low impedance).
  //for fine tuning, wait until a millisecond tick.
  digitalWrite(this->PinLED, HIGH); //set status LED to on.
  this->TimeStart=millis()+2;
  while(millis() < this->TimeStart) {} //wait for the tick.
  digitalWrite(this->PinChargeResistor, HIGH); //set pin HIGH for capacitor charging.
  this->TimeStart = millis(); //start the clock.
  while(analogRead(this->PinADCCapacitor) < this->CapacitorADCValue) {}
  this->TimeElapsed = millis() - this->TimeStart; //stop the clock.
  digitalWrite(this->PinLED, LOW); //set status LED to off.
  this->Discharge(this->CapacitorADCValue); //drain and set up for next run.
} //end Time();



//Dicharge the capacitor.
//adcValue less than 100 will trigger quick drain mode.
void CapacitanceMeter::Discharge(unsigned int adcValue) {
  if(this->TimeElapsed < 100) {adcValue=7;} //speed things up on low value reads.
  pinMode(this->PinChargeResistor, OUTPUT); //set to output (low impedance). Paranoia: this should already be done.
  digitalWrite(this->PinChargeResistor, LOW); //set pin to LOW (low impedance).
  pinMode(this->PinDischargeResistor, OUTPUT); //set discharge pin to output (low impedance).
  digitalWrite(this->PinDischargeResistor, LOW); //set discharge pin LOW (low impedance).
  while(analogRead(this->PinADCCapacitor) > 0) {} //wait until capacitor is completely discharged.
  adcValue<100?delay(100):delay(1000); //hold discharge a tad longer (also doesn't bash serial output).
  pinMode(this->PinDischargeResistor, INPUT); //set discharge pin back to input (high impedance).
  pinMode(this->PinChargeResistor, INPUT); //set to input (high impedance) to not affect the other pins.
} //end Discharge();



//Print out the results to the serial port.
void CapacitanceMeter::SerialPrint() {
  if(this->milliFarads==0.0) {return;} //don't print out empty values.
  Serial.print(this->ChargeResistorValue, 0);
  Serial.print("r ");
  Serial.print(this->TimeElapsed); //print the value to serial port
  Serial.print("mS ");
  if(this->milliFarads < 1.0e-3) { //print the value out as nanoFarads.
    Serial.print((this->milliFarads*1.0e6), 2); //print the value to serial port
    Serial.println(" nanoFarads"); //print units and carriage return
    }
  else if(this->milliFarads < 1.0) { //print the value out as microFarads.
    Serial.print((this->milliFarads*1.0e3), 3); //print the value to serial port
    Serial.println(" microFarads"); //print units and carriage return
    }
  else { //print the value out as milliFarads.
    Serial.print(this->milliFarads, 3); //print the value to serial port
    Serial.println(" milliFarads"); //print units and carriage return
    }
} //end SerialPrint();



//Get capacitance using resistance values on pin.
void CapacitanceMeter::Get() {
  //get the time for the capacitor.
  this->Time();
  //Formula: Farads = Seconds / Resistance.
  //Since seconds (to 63.2% charge) is milliseconds (by the millis() clock), results are milliFarads.
  this->milliFarads = (double)this->TimeElapsed / this->ChargeResistorValue;
  //TimeStart is used to see if something happened in the middle of all this, so zero it when finished.
  TimeStart=0;
} //end Get();

//end of CapacitanceMeter() class.
