// CapacitanceMeter.h definition of class. For Arduino 0018 or newer.
// CapacitanceMeter class to consolidate various variables and functions.
// (c)20110312 by MODAT7 under GNU GPLv1. Original single code by Paul Badger.


/* Notes.

This class uses the voltage rise time through a resistor to calculate the value of a capacitor.
This class was derived from Paul Badger's original 2008 example and implements some of his
suggestions here and in the main program body. This class was written because I needed to test
some large salvaged capacitors and my meter doesn't go over 20uF (it seems few meters go high).
With multiple objects tied together with some glue code, an auto-ranging meter may be developed.

With a properly metered charge resistor and measurment times over 200mS, error tolerance is about
3%. Note that the ESR of a capacitor can skew the reading some. A function has been included to
show the percentage between two different resistors to give a very rough idea of the ESR. A low
ESR capacitor should be less than 1% different for any resistor. 1% tolerance (or better) charge
resistors do not have to be used so long as the resistor in question can be accurately metered.
Calibrating a resistor value in the program against a known capacitor isn't always recommended
because of ESR variations skewing the results (several high quality 1% and very low ESR
capacitors "may" work out, though).

When choosing a charge resistor, do not have it small enough to pull more than 10mA from the
AVR pin (500ohm at 5v). While an AVR pin can source up to 40mA, pushing it to the limits does
not guarantee that 40mA will always be available. If other pins are also sourcing current,
this value will be unpredictably lower.

Do not overload the charge resistor pins with LED's. This will form voltage dividers and also
suck away needed current. In this class, I gave the LED its own pin specifically to avoid this
problem.

This class can measure capacitors into the milliFarad range, but be warned that measurment times
will greatly rise. Be patient with them.

Smaller capacitors need larger resistors. About a 10meg-ohm resistor would be needed to measure
1nF accurately. The issue with larger resistors is that they start becoming the values of the
AVR pins when they are in INPUT mode. This can start skewing the reading badly if multiple pins
are tied together and share an ADC pin (such as in an auto-ranging configuration).

This class has been specifically broken up into modular pieces to allow for different
functionality and reporting in the future. For version 2a I added code for a 16x2 LCD for
standalone reporting, instead of using serial alone. An LCD with more than 2 lines will not have
the extra lines used. An LCD with more than 16 characters across may have them used for large
value reporting.

Some will argue that as a class, this is somewhat inefficient with space when making an
auto-ranging meter. This is true to a point, but the overlapping variables aren't that many and
the current AVR's have plenty of space to handle this. In this case, clarity of operation is a
better choice. Keeping track of the extra time and Farads also allows for multiple comparisons
of the same capacitor by different resistors. I've already added an ESR percent difference based
on this. That could be expanded on with other possible functions. Since the ATmega328 has 6
ADC's, it would also be trivial to set up multiple independent meters using this class.

The rest of the source code is excessively commented because of my memory problems. Read through
it carefully to learn how everything works and to do your own custom setup.

*/

#ifndef CAPACITANCEMETER_H
#define CAPACITANCEMETER_H

//Bring in the Arduino stuff since it gets called.
#include <WProgram.h>

// CapacitanceMeter class to consolidate various variables and functions.
class CapacitanceMeter {

public:
//note: floats and doubles are both 4 bytes in avr-gcc.
unsigned char PinChargeResistor; //pin number on the Arduino board the charge resistor is connected to.
unsigned char PinADCCapacitor; //ADC pin number on the Arduino board the capacitor is connected to.
unsigned char PinDischargeResistor; //pin number on the Arduino board the discharge resistor is connected to.
unsigned char PinLED; //pin number on the Arduino board the LED is connected to.
unsigned int CapacitorADCValue; //count up value to stop reading the ADC and stop the clock. This will vary depending on ADC bits.
double ChargeResistorValue; //value of the charge resistor.

unsigned long TimeStart; //keep track of millis() start time.
unsigned long TimeElapsed; //milliseconds to reach CapacitorADCValue.
double milliFarads; //calculated value of capacitor.

//CapacitanceMeter(); //Constructor, removed for size.
void SetUp(unsigned char, unsigned char, unsigned char, unsigned char, unsigned int, double);
void Check();
void Time();
void Discharge(unsigned int);
void SerialPrint();
void Get();
};

#endif

//end of CapacitanceMeter.h
